package org.fuseri.moduleexercise.question;

import com.fasterxml.jackson.databind.ObjectMapper;
import jakarta.persistence.EntityNotFoundException;
import org.fuseri.model.dto.exercise.AnswerDto;
import org.fuseri.model.dto.exercise.AnswerInQuestionCreateDto;
import org.fuseri.model.dto.exercise.ExerciseCreateDto;
import org.fuseri.model.dto.exercise.QuestionCreateDto;
import org.fuseri.model.dto.exercise.QuestionDto;
import org.fuseri.model.dto.exercise.QuestionUpdateDto;
import org.fuseri.moduleexercise.exercise.Exercise;
import org.junit.jupiter.api.AfterEach;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentMatcher;
import org.mockito.ArgumentMatchers;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.boot.test.autoconfigure.web.servlet.AutoConfigureMockMvc;
import org.springframework.boot.test.context.SpringBootTest;
import org.springframework.boot.test.mock.mockito.MockBean;
import org.springframework.http.MediaType;
import org.springframework.http.ResponseEntity;
import org.springframework.test.web.servlet.MockMvc;
import org.springframework.test.web.servlet.ResultActions;

import static org.hamcrest.Matchers.is;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.when;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.put;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.jsonPath;
import static org.springframework.test.web.servlet.result.MockMvcResultMatchers.status;

import java.util.ArrayList;
import java.util.List;

import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.delete;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.get;
import static org.springframework.test.web.servlet.request.MockMvcRequestBuilders.post;


@SpringBootTest
@AutoConfigureMockMvc
public class QuestionTest {

    @Autowired
    ObjectMapper objectMapper;
    @Autowired
    private MockMvc mockMvc;

    @MockBean
    QuestionFacade facade;

    private QuestionDto qston;

    public static String asJsonString(final Object obj) {
        try {
            return new ObjectMapper().writeValueAsString(obj);
        } catch (Exception e) {
            throw new RuntimeException(e);
        }
    }

    private void createExercise() {
        var postExercise = new ExerciseCreateDto("idioms", "exercise on basic idioms", 2, 0L);
        var postExercise2 = new ExerciseCreateDto("riddles", "simple english riddles", 2, 0L);
        try {
            mockMvc.perform(post("/exercises").content(asJsonString(postExercise)).contentType(MediaType.APPLICATION_JSON));
            mockMvc.perform(post("/exercises").content(asJsonString(postExercise2)).contentType(MediaType.APPLICATION_JSON));

        } catch (Exception e) {
            assert (false);
        }
    }

    @BeforeEach
    void init() {
//        createExercise();

        qston = new QuestionDto("\"what is the meaning of: costs an arm and leg\"",1,new ArrayList<>());
        long id = 2;

        var question = new QuestionCreateDto("this statement is false", id, List.of(new AnswerInQuestionCreateDto("dis a logical paradox", true)));
        var question2 = new QuestionCreateDto("What month of the year has 28 days?", id, List.of(new AnswerInQuestionCreateDto("February", false), new AnswerInQuestionCreateDto("All of them", true)));
        ResultActions posted = null;
    }


    @Test
    void testCreateQuestion() throws Exception {
        var exerciseId = 1L;
        var question = new QuestionCreateDto("what is the meaning of: costs an arm and leg", exerciseId, List.of(new AnswerInQuestionCreateDto("dis very expencive", true), new AnswerInQuestionCreateDto("FMA refference", false)));
        when(facade.create(ArgumentMatchers.isA(QuestionCreateDto.class))).thenReturn(qston);
        var posted = mockMvc.perform(post("/questions").content(asJsonString(question)).contentType(MediaType.APPLICATION_JSON));

        posted.andExpect(status().isCreated())
                .andExpect(jsonPath("$.text", is(qston.getText())))
                .andExpect(jsonPath("$.exerciseId", is((int)qston.getExerciseId())));

    }

    @Test
    void testCreateQuestionEmptyQuestions() throws Exception {
        var prompt = """
                {
                "text": "this is a question without exercise Id",
                "exerciseId": 1,
                "answers": []
                }
                """;

        var posted = mockMvc.perform(post("/questions").content(prompt).contentType(MediaType.APPLICATION_JSON));

        posted.andExpect(status().isCreated());
    }


    @Test
    void testCreateQuestionEmptyField() throws Exception {
        var exerciseId = 1L;
        var question = new QuestionCreateDto("", exerciseId, List.of(new AnswerInQuestionCreateDto("dis very expencive", true), new AnswerInQuestionCreateDto("FMA refference", false)));
        var posted = mockMvc.perform(post("/questions").content(asJsonString(question)).contentType(MediaType.APPLICATION_JSON));

        posted.andExpect(status().is4xxClientError());
    }

    @Test
    void testCreateQuestionMissingField() throws Exception {
        var prompt = """
                {
                "text": "this is a question without exercise Id,
                "answers" : []
                }
                """;

        var posted = mockMvc.perform(post("/questions").content(prompt).contentType(MediaType.APPLICATION_JSON));

        posted.andExpect(status().is4xxClientError());
    }

    private QuestionDto createQuestion(long id) throws Exception {
        var question = new QuestionCreateDto("this statement is false", id, List.of(new AnswerInQuestionCreateDto("dis a logical paradox", true)));
        var posted = mockMvc.perform(post("/questions").content(asJsonString(question)).contentType(MediaType.APPLICATION_JSON));
        var cont = posted.andReturn().getResponse().getContentAsString();
        var res = objectMapper.readValue(cont, QuestionDto.class);
        return res;
    }


    @Test
    void getQuestion() throws Exception {
        var question = new QuestionDto("this statement is false",1L,new ArrayList<>());
        when(facade.find(1)).thenReturn(question);
        var gets = mockMvc.perform(get("/questions/1"));
        gets.andExpect(status().isOk()).andExpect(jsonPath("$.text", is("this statement is false")));
    }

    @Test
    void getQuestionNotFound() throws Exception {
        when(facade.find(9999)).thenThrow(new EntityNotFoundException());
        var gets = mockMvc.perform(get("/questions/9999"));
        gets.andExpect(status().isNotFound());
    }


    @Test
    void getAnswer() throws Exception {


        var sss = List.of(new AnswerDto("February", false), new AnswerDto("All of them", true));
        when(facade.getQuestionAnswers(2)).thenReturn(sss);
        var gets = mockMvc.perform(get("/questions/2/answers"));
        gets.andExpect(status().isOk())
                .andExpect(jsonPath("$[0].text", is("February")))
                .andExpect(jsonPath("$[1].text", is("All of them")));


    }

    @Test
    void TestUpdate() throws Exception {
        var updated = """
                {
                  "text": "wat a paradox?",
                  "exerciseId": "1"
                }
                """;

        var question = new QuestionDto("wat a paradox?",1,new ArrayList<>());

        when(facade.patchUpdate(ArgumentMatchers.eq(1), ArgumentMatchers.isA(QuestionUpdateDto.class))).thenReturn(question);
        var gets = mockMvc.perform(put("/questions/1").content(updated).contentType(MediaType.APPLICATION_JSON));

        gets.andExpect(status().isOk());
    }

@Test
    void deleteExisting() {
    try {
        mockMvc.perform(delete("/questions/1")).andExpect(status().isNoContent());
    } catch (Exception e) {
        throw new RuntimeException(e);
    }
}

}
